import * as THREE from 'https://threejsfundamentals.org/threejs/resources/threejs/r112/build/three.module.js';
import {OrbitControls} from 'https://threejsfundamentals.org/threejs/resources/threejs/r112/examples/jsm/controls/OrbitControls.js';
import {GLTFLoader} from 'https://threejsfundamentals.org/threejs/resources/threejs/r112/examples/jsm/loaders/GLTFLoader.js';

//MelonEngine 0.1 - melonking.net

//Settings
const PRINT_CONSOLE_ROWS = 16;
const MIN_TIME_TO_START = 60;

//HTML Elements
export const canvas = document.querySelector('#plot-canvas');
export const console = document.querySelector('#plot-console');

//Other Stuff
export const modelLoader = new GLTFLoader();
export const audioLoader = new THREE.AudioLoader();
export const clock = new THREE.Clock();
export const raycaster = new THREE.Raycaster();
export const mouse = new THREE.Vector2();
export const listener = new THREE.AudioListener();

export const renderer = getStandardRenderer();
export const camera = getStandardCamera();

export const actionObjects = [];
export const consoleHistory = [];

var mixers = []; //Animation Array

//Audio
var mouseMoveSound;
var globalAudios = [];
var localAudios = [];

var extraInit = function(){};
var extraStart = function(){};
var extraLoop = function(){};

var gameClock = 0;
var isReady = false;
var scene, controls;

export function startErUp(pExtraInit, pExtraStart, pExtraLoop)
{
	extraInit = pExtraInit;
	extraStart = pExtraStart;
	extraLoop = pExtraLoop;
	init();
	masterLoop();
}

//Basic setup and loading
function init() 
{
	//Setup
	controls = getStandardControls();
	camera.add( listener );
	
	mouseMoveSound = loadGlobalAudio( 'feet-short.mp3', false ); //TEMP HACK FOR GAME JAM REMOVE
	
	scene = new THREE.Scene();
	scene.background = new THREE.Color( 0xfffefc );
	scene.fog = new THREE.Fog( 0xfffefc, 0, 1000 );
	
	//Events
	document.addEventListener( 'mousedown', onDocumentMouseDown, false );
	window.addEventListener( 'resize', onWindowResize, false );
	
	extraInit.call();
}

//Called once when all loading is done
function start()
{
	if(gameClock < MIN_TIME_TO_START || isReady || scene.getObjectByName( 'Ground' ) == undefined) {return;} //Scene is setup or not ready yet
	
	extraStart.call();
	
	isReady = true;
}

//Game loop
function masterLoop()
{	
	gameClock++;
	
	start();
	var delta = clock.getDelta();
	
	//Camera logic
	raycaster.setFromCamera( new THREE.Vector2(), camera );
	var inFrontOfCamera = new THREE.Vector3();
	raycaster.ray.at( 0.1, inFrontOfCamera );
	controls.target = inFrontOfCamera;
	
	//Animation support
	for(var m=0 ; m<=getLatestMixer() ; m++)
	{
		mixers[m].update(delta);
	}
	
	//Custom loop code
	extraLoop.call();
	
	//Standard Loop Logic
	controls.update();
	renderer.render( scene, camera );
	requestAnimationFrame( masterLoop );
}

//+++ Setup helpers +++

export function getStandardCamera()
{
	var camera = new THREE.PerspectiveCamera( 70, canvas.clientWidth / canvas.clientHeight, 0.1, 1000 );
	camera.position.y = 1;
	return camera;
}

export function getStandardRenderer()
{
	var renderer = new THREE.WebGLRenderer( { canvas, antialias: true } );
	renderer.setPixelRatio( window.devicePixelRatio );
	renderer.setSize( canvas.clientWidth, canvas.clientHeight );
	renderer.physicallyCorrectLights = false;
	return renderer;
}

export function getStandardControls()
{
	var controls = new OrbitControls( camera, document.body );
	controls.enableDamping = true;
	controls.enableZoom = false;
	controls.enableKeys = true;
	controls.dampingFactor = 0.07;
	controls.screenSpacePanning = false;
	controls.minDistance = 0;
	controls.maxDistance = 1;
	controls.maxPolarAngle = Math.PI / 1.1;
	controls.panSpeed = 130;
	controls.keyPanSpeed = 10000;
	return controls;
}

//+++ Loaders +++

export function loadObject(name, scale, rename='')
{
	modelLoader.load( '3d/models/'+name+'.gltf', (gltf) => {

		var model = gltf.scene;
		var animations = gltf.animations;
		
		model.scale.set(scale,scale,scale);
		model.children[0].name = rename;
		
		scene.add( model );
		
		//Animation Support
		if(gltf.animations.length > 0)
		{
			mixers.push(new THREE.AnimationMixer( model ));
			var action = mixers[getLatestMixer()].clipAction( animations[0] );
			action.play();
		}
	});
}

export function loadGlobalAudio(name, isLoop=true, volume=0.5)
{
	var audioClip = new THREE.Audio( listener );
		audioLoader.load( 'audio/'+name, function( buffer ) {
		audioClip.setBuffer( buffer );
		audioClip.setLoop( isLoop );
		audioClip.setVolume( volume );
	});
	return audioClip;
}

export function loadLocalAudio(name, distance=20, isLoop=true, volume=0.5)
{
	var audioClip = new THREE.PositionalAudio( listener );
	audioLoader.load( 'audio/'+name, function( buffer ) {
		audioClip.setBuffer( buffer );
		audioClip.setRefDistance( distance );
		audioClip.setLoop( isLoop );
		audioClip.setVolume( volume );
	});
	return audioClip;
}

//+++ Runtime functions +++

export function getActiveScene()
{
	return scene;
}

export function getLatestMixer()
{
	return mixers.length-1;
}

export function getObjectByName(name )
{
	return scene.getObjectByName( name );
}

export function getDelta()
{
	return clock.getDelta() * 1000;
}

//+++ Main system functionality +++

//Window resizing logic
function onWindowResize()
{
	camera.aspect = window.innerWidth / window.innerHeight;
	camera.updateProjectionMatrix();
	renderer.setSize( window.innerWidth, window.innerHeight );
}

//Object clicking logic
function onDocumentMouseDown( event ) 
{ 	
	if(event.button == 2) { if(!mouseMoveSound.isPlaying) { mouseMoveSound.play() } } //TEMP HACK FOR GAME JAM REMOVE
	
	if(event.button != 0) { return; } //Left click only
	if(actionObjects.length < 1) { return; } //No actions, return
	
	mouse.x = ( event.clientX / renderer.domElement.clientWidth ) * 2 - 1;
	mouse.y = - ( event.clientY / renderer.domElement.clientHeight ) * 2 + 1;
	
	raycaster.setFromCamera( mouse, camera );
	
	var intersects = raycaster.intersectObjects( actionObjects, true );
	
	if(intersects.length < 1) {return;} //Nothing clicked, return
	
	var action = intersects[0].object.MELONAction;
	if(action == undefined) { action = intersects[0].object.parent.MELONAction } //Bad fix for multi depth objects REDO LATER
	
	var oneTime = intersects[0].object.MELONActionOnetime;
	if(oneTime == 2) { return; } //Action is used
	else if(oneTime == 1) { intersects[0].object.MELONActionOnetime = 2; }
	
	action.call();
}

//+++ Standard Interactions +++

export function registerAction(objectName, action, oneTime=false)
{
	var object = scene.getObjectByName( objectName );
	if(object == undefined) {return;}
	
	//Crappy onetime system
	object.MELONActionOnetime = 0; //0 no, 1 not used, 2 used
	if(oneTime) {object.MELONActionOnetime = 1;}
	
	object.MELONAction = action;
	
	actionObjects.push(object);
}

export function printToConsole(text)
{
	consoleHistory.push(text);
	var lastIndex = consoleHistory.length - PRINT_CONSOLE_ROWS;
	var consoleText = '';
	var consoleFlop = true;
	
	for(var i=0 ; i<PRINT_CONSOLE_ROWS ; i++)
	{
		var entry = consoleHistory[lastIndex+i];
		if(entry == undefined || entry == '') {continue;}
		if(consoleFlop) { consoleText += '<br/><div class="console-flop">' + entry + '</div>'; consoleFlop = false; }
					else { consoleText += '<br/>' + entry; consoleFlop = true; }
	}
	console.innerHTML = consoleText;
}

//+++ Game Helper Functions +++
export function getRandomInt(min, max)
{
	return Math.floor((Math.random() * max) + min);
}